
# coding: utf-8

# In[1]:


from vis.utils import utils
from keras import activations
from keras.applications import VGG16,Xception,InceptionV3
from keras import backend as K 
from keras.models import Model


# In[2]:


# Build the VGG16 network with ImageNet weights
# if K.image_data_format() == 'channels_first':
#     input_shape = (3, 299, 299)
# else:
#     input_shape = (299, 299, 3)
K.set_image_data_format('channels_last')
K.set_image_dim_ordering('tf')
model = InceptionV3(weights='imagenet', include_top=True,input_shape=(299,299,3))
# Utility to search for layer index by name. 
# Alternatively we can specify this as -1 since it corresponds to the last layer.
layer_idx = utils.find_layer_idx(model, 'predictions')

# Swap softmax with linear
model.layers[layer_idx].activation = activations.linear
model = utils.apply_modifications(model)


# In[3]:


# for i,layer in enumerate(model.layers):
#     print(i,layer.name)


# In[4]:


from vis.utils import utils
from matplotlib import pyplot as plt
get_ipython().run_line_magic('matplotlib', 'inline')
plt.rcParams['figure.figsize'] = (18, 6)

img1 = utils.load_img('./outzel/ouzel1.jpg', target_size=(299, 299))
img2 = utils.load_img('./outzel/ouzel2.jpg', target_size=(299, 299))
print(img1.shape)
f, ax = plt.subplots(1, 2)
ax[0].imshow(img1)
ax[1].imshow(img2)


# In[5]:


from vis.visualization import visualize_saliency, overlay
from vis.utils import utils
from keras import activations

# Utility to search for layer index by name. 
# Alternatively we can specify this as -1 since it corresponds to the last layer.
layer_idx = utils.find_layer_idx(model, 'predictions')

f, ax = plt.subplots(1, 2)
for i, img in enumerate([img1, img2]):    
    # 20 is the imagenet index corresponding to `ouzel`
    grads = visualize_saliency(model, layer_idx, filter_indices=20, seed_input=img)
    
    # visualize grads as heatmap
    ax[i].imshow(grads, cmap='jet')


# In[6]:


for modifier in ['guided', 'relu']:
    plt.figure()
    f, ax = plt.subplots(1, 2)
    plt.suptitle(modifier)
    for i, img in enumerate([img1, img2]):    
        # 20 is the imagenet index corresponding to `ouzel`
        grads = visualize_saliency(model, layer_idx, filter_indices=20, 
                                   seed_input=img, backprop_modifier=modifier)
        # Lets overlay the heatmap onto original image.    
        ax[i].imshow(grads, cmap='jet')


# In[7]:


# model.summary()


# In[8]:


import numpy as np
import matplotlib.cm as cm
from vis.visualization import visualize_cam

for modifier in [None, 'guided', 'relu']:
    plt.figure()
    f, ax = plt.subplots(1, 2)
    plt.suptitle("vanilla" if modifier is None else modifier)
    for i, img in enumerate([img1, img2]):    
        # 20 is the imagenet index corresponding to `ouzel`
        grads = visualize_cam(model, layer_idx, filter_indices=20, 
                              seed_input=img, backprop_modifier=modifier)        
        # Lets overlay the heatmap onto original image.    
        jet_heatmap = np.uint8(cm.jet(grads) * 255)[:, : , :, 3]
        ax[i].imshow(overlay(grads, img))

