
# coding: utf-8

# In[1]:


import os
print(len(os.listdir('./CUB_200_2011/299/train/')))


# In[2]:


import os
import cv2
import matplotlib.pyplot as plt
from keras.utils import  plot_model
from keras.preprocessing.image import  ImageDataGenerator
from keras.layers import Dense,Flatten,GlobalAveragePooling2D,Dropout,Input
from keras.models import Model,load_model
from keras.applications.vgg19 import VGG19
from keras.applications.xception import Xception
import keras.callbacks as kcallbacks
import keras
from keras.applications.inception_v3 import InceptionV3
from keras.preprocessing.image import load_img,img_to_array
import math
import random
from PIL import Image
import numpy as  np
from keras.callbacks import CSVLogger
from keras.applications.resnet50 import ResNet50


# In[3]:


###########配置的文件####################
TARGET_SIZE=(299,299)
INPUT_SIZE=(299,299,3)
BATCHSIZE=64
##################################
train_datagen = ImageDataGenerator(rescale=1./255)

test_datagen = ImageDataGenerator(rescale=1./255)

train_generator = train_datagen.flow_from_directory(
        './CUB_200_2011/299/train/',
        target_size=TARGET_SIZE,
        batch_size=BATCHSIZE,
        class_mode='categorical')
validation_generator = test_datagen.flow_from_directory(
        './CUB_200_2011/299/valid/',
        target_size=TARGET_SIZE,
        shuffle=False,
        batch_size=BATCHSIZE,
        class_mode='categorical')


# In[4]:


def model_inception(input_shape,
                    num_class,
                    epochs,
                    savepath='./inc_model.h5'):
    model_inception = InceptionV3(
        include_top=False, weights='imagenet', input_shape=input_shape)
    print(len(model_inception.layers))
    for layer in model_inception.layers[:200]:
        layer.trainable = False
    for layer in model_inception.layers[200:]:
        layer.trainable = True
    # 平均池化之后接全连接层
    model = GlobalAveragePooling2D()(model_inception.output)
    model = Dense(num_class, activation='softmax')(model)
    model_inception = Model(model_inception.input, model, name='inception')
    #定义一下优化的方法
    opt = keras.optimizers.Adam(
        lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    #compile一下模型
    model_inception.compile(
        loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
    #train model
    earlyStopping = kcallbacks.EarlyStopping(
        monitor='val_acc', patience=5, verbose=1, mode='auto')
    csv_logger = CSVLogger('inception_log.csv')
    saveBestModel = kcallbacks.ModelCheckpoint(
        filepath=savepath,
        monitor='val_acc',
        verbose=1,
        save_best_only=True,
        mode='auto')
    hist = model_inception.fit_generator(
        train_generator,
        steps_per_epoch=len(train_generator),
        epochs=epochs,
        validation_data=validation_generator,
        validation_steps=len(validation_generator),
        callbacks=[earlyStopping, saveBestModel,csv_logger])
    model_inception = load_model(savepath)
    return model_inception
model_inception = model_inception(INPUT_SIZE, num_class=200, epochs=20)


# # 测试一下xception

# In[9]:


def model_xception(input_shape,num_class,epochs,savepath='./xcep_model.h5'):
#     model_xception = Xception(include_top=False, weights='imagenet',input_shape=input_shape)
#     print(len(model_xception.layers))
    base_model = Xception(include_top=True, weights='imagenet', input_tensor=None, input_shape=None)
    plot_model(base_model, to_file='xception_model.png')
    base_model.layers.pop()
    base_model.outputs = [base_model.layers[-1].output]
    base_model.layers[-1].outbound_nodes = []
    base_model.output_layers = [base_model.layers[-1]]
    feature = base_model
    img = Input(shape=(299, 299, 3), name='img')
    feature1 = feature(img)
    for layer in base_model.layers[:105]:  
        layer.trainable = False 
    for layer in base_model.layers[105:]:  
        layer.trainable = True
#     x=model_xception.output
    category_predict = Dense(num_class, activation='softmax', name='category_predict')(
        Dropout(0.5)(feature1)
    )
    model=Model(img,category_predict,name='xception')
    plot_model(model,to_file='./xcpetion.png')
    #定义一下优化的方法
    opt = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    #compile一下模型
    model.compile(loss='categorical_crossentropy',optimizer=opt,metrics=['accuracy'])
    #train model
    earlyStopping=kcallbacks.EarlyStopping(monitor='val_acc', patience=5, verbose=1, mode='auto')
    saveBestModel = kcallbacks.ModelCheckpoint(filepath=savepath, monitor='val_acc', verbose=1, save_best_only=True, mode='auto')
    hist=model.fit_generator(
        train_generator,
        steps_per_epoch=len(train_generator),
        epochs=epochs,
        validation_data=validation_generator,
        validation_steps=len(validation_generator),
        callbacks=[earlyStopping,saveBestModel]
    )
    with open('log.txt','w') as f:
        f.write(str(hist.history))
    model=load_model(savepath)
    return model_xception


# In[10]:


model_inception=model_xception(INPUT_SIZE,num_class=200,epochs=20)


# # Resnet-50

# In[13]:


def model_res(input_shape,
                    num_class,
                    epochs,
                    savepath='./resnet-50.h5'):
    model_inception = ResNet50(
        include_top=False, weights='imagenet', input_shape=input_shape)
    print(len(model_inception.layers))
    for layer in model_inception.layers[:100]:
        layer.trainable = False
    for layer in model_inception.layers[100:]:
        layer.trainable = True
    # 平均池化之后接全连接层
    model = GlobalAveragePooling2D()(model_inception.output)
    model = Dense(num_class, activation='softmax')(model)
    model_inception = Model(model_inception.input, model, name='inception')
    #定义一下优化的方法
    opt = keras.optimizers.Adam(
        lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    #compile一下模型
    model_inception.compile(
        loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
    #train model
    earlyStopping = kcallbacks.EarlyStopping(
        monitor='val_acc', patience=5, verbose=1, mode='auto')
    csv_logger = CSVLogger('res_log.csv')
    saveBestModel = kcallbacks.ModelCheckpoint(
        filepath=savepath,
        monitor='val_acc',
        verbose=1,
        save_best_only=True,
        mode='auto')
    hist = model_inception.fit_generator(
        train_generator,
        steps_per_epoch=len(train_generator),
        epochs=epochs,
        validation_data=validation_generator,
        validation_steps=len(validation_generator),
        callbacks=[earlyStopping, saveBestModel,csv_logger])
    model_inception = load_model(savepath)
    return model_inception
modelres = model_res(INPUT_SIZE, num_class=200, epochs=20)

