
# coding: utf-8

# In[1]:


import os
import numpy as np
from keras import Input
from keras.applications import Xception, InceptionV3,VGG16
from keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau
from keras.layers import Dense, Dropout, concatenate, maximum,MaxPooling2D,AveragePooling2D,Conv2D,BatchNormalization
from keras.layers.convolutional import  SeparableConv2D
from keras.models import Model,load_model
from keras.optimizers import SGD
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import plot_model,to_categorical
from keras.layers.core import Lambda
import os
import keras
from collections import OrderedDict
from sklearn.preprocessing import LabelEncoder
import random
from keras.preprocessing import image
from numpy.random import randint, shuffle, choice
import keras.callbacks as kcallbacks
import cv2
import math


# In[2]:


def resize_image(x, size_target=None, flg_keep_aspect=False, rate_scale=1.0, flg_random_scale=False):

    # convert to numpy array
    if not isinstance(x, np.ndarray):
        img = np.asarray(x)
    else:
        img = x

    # calculate resize coefficients
    if len(img.shape) == 4:
        _o, size_height_img, size_width_img, _c , = img.shape
        img = img[0]
    elif len(img.shape) == 3:
        size_height_img, size_width_img, _c , = img.shape

    if len(size_target) == 1:
        size_heigth_target = size_target
        size_width_target = size_target
    if len(size_target) == 2:
        size_heigth_target = size_target[0]
        size_width_target = size_target[1]
    if size_target == None:
        size_heigth_target = size_height_img * rate_scale 
        size_width_target = size_width_img * rate_scale 

    coef_height = 1
    coef_width = 1
    if size_height_img < size_heigth_target :
        coef_height = size_heigth_target / size_height_img
    if size_width_img < size_width_target :
        coef_width = size_width_target / size_width_img

    # calculate coeffieient to match small size to target size
    ## scale coefficient if specified
    low_scale = rate_scale
    if flg_random_scale:
        low_scale = 1.0
    coef_max = max(coef_height, coef_width) * np.random.uniform(low=low_scale, high=rate_scale)

    # resize image
    size_height_resize = math.ceil(size_height_img*coef_max)
    size_width_resize = math.ceil(size_width_img*coef_max)

    # method_interpolation = cv2.INTER_LINEAR
    method_interpolation = cv2.INTER_CUBIC
    # method_interpolation = cv2.INTER_NEAREST

    if flg_keep_aspect:
        img_resized = cv2.resize(
                            img
                            , dsize=(size_width_resize, size_height_resize)
                            , interpolation=method_interpolation
                        )
    else:
        img_resized = cv2.resize(
                            img
                            , dsize=(
                                int(size_width_target*np.random.uniform(low=low_scale, high=rate_scale))
                                ,int(size_heigth_target*np.random.uniform(low=low_scale, high=rate_scale))
                            )
                            , interpolation=method_interpolation
                        )
    return img_resized

def resize_images(images, **kwargs):
    max_images = len(images)
    for i in range(max_images):
        images[i] = resize_image(images[i], **kwargs)
    return images


# In[3]:


# crop image at the center
def center_crop_image(x, size_target=(448,448)):

    # convert to numpy array
    if not isinstance(x, np.ndarray):
        img = np.asarray(x)
    else:
        img = x

    # set size
    if len(size_target) == 1:
        size_heigth_target = size_target
        size_width_target = size_target
    if len(size_target) == 2:
        size_heigth_target = size_target[0]
        size_width_target = size_target[1]

    if len(img.shape) == 4:
        _o, size_height_img, size_width_img, _c , = img.shape
        img = img[0]
    elif len(img.shape) == 3:
        size_height_img, size_width_img, _c , = img.shape

    # crop image
    h_start = int((size_height_img - size_heigth_target) / 2)
    w_start = int((size_width_img - size_width_target) / 2)
    img_cropped = img[h_start:h_start+size_heigth_target, w_start:w_start+size_width_target, :]

    return img_cropped


# crop image of fixed-size from random point of top-left corner
def random_crop_image(x, size_target=(448,448)):

    # convert to numpy array
    if not isinstance(x, np.ndarray):
        img = np.asarray(x)
    else:
        img = x

    # set size
    if len(size_target) == 1:
        size_heigth_target = size_target
        size_width_target = size_target
    if len(size_target) == 2:
        size_heigth_target = size_target[0]
        size_width_target = size_target[1]

    if len(img.shape) == 4:
        _o, size_height_img, size_width_img, _c , = img.shape
        img = img[0]
    elif len(img.shape) == 3:
        size_height_img, size_width_img, _c , = img.shape

    # crop image
    margin_h = (size_height_img - size_heigth_target)
    margin_w = (size_width_img - size_width_target)
    h_start = 0 
    w_start = 0
    if margin_h != 0:
        h_start = np.random.randint(low=0, high=margin_h)
    if margin_w != 0:
        w_start = np.random.randint(low=0, high=margin_w) 
    img_cropped = img[h_start:h_start+size_heigth_target, w_start:w_start+size_width_target, :]

    return img_cropped


# flip image horizontally 
def horizontal_flip_image(x):
    
    if np.random.random() >= 0.5:
        return x[:,::-1,:]
    else:
        return x


# feature-wise normalization
def normalize_image(x, mean=(0., 0., 0.), std=(1.0, 1.0, 1.0)) :
    
    x = np.asarray(x, dtype=np.float32)
    x=x/255.
#     if len(x.shape) == 4:
#         for dim in range(3):
#             x[:,:,:,dim] = ( x[:,:,:,dim] - mean[dim] ) / std[dim]
#     if len(x.shape) == 3:
#         for dim in range(3):
#             x[:,:,dim] = ( x[:,:,dim] - mean[dim] ) / std[dim]

    return x


def check_images(images):
    fig=plt.figure(figsize=(8, 8))
    columns = 3
    rows = 3
    for i in range(1, columns*rows+1):
        fig.add_subplot(rows, columns, i)
        plt.imshow(images[i-1])
    plt.show()


# In[4]:


###########配置的文件####################
TARGET_SIZE=(448,448)
INPUT_SIZE=(448,448,3)
BATCHSIZE=15
EPOCH=20
 # set preprocessing functions       
func_train = lambda x :normalize_image(
                            random_crop_image(
                                horizontal_flip_image(
                                    resize_image(x, size_target=TARGET_SIZE, flg_keep_aspect=True)
                                )
                            )
#                             ,mean=[123.82988033, 127.3509729, 110.25606303]
                        )
func_valid = lambda x :normalize_image(
                            center_crop_image(
                                resize_image(x, size_target=TARGET_SIZE, flg_keep_aspect=True)
                            )
#                             ,mean=[123.82988033, 127.3509729, 110.25606303]
                        )
train_datagen = ImageDataGenerator(
                                   preprocessing_function=func_train
                                  )
test_datagen = ImageDataGenerator(preprocessing_function=func_valid)



# train_datagen = ImageDataGenerator(rescale=1./255)
# train_datagen = ImageDataGenerator(
#                                 horizontal_flip=True,
#                                 vertical_flip=True,
#                                 rescale=1./255
#                                   )
# test_datagen = ImageDataGenerator(rescale=1./255)

train_generator = train_datagen.flow_from_directory(
        './CUB_200_2011/splitted/train/',
        target_size=TARGET_SIZE,
        batch_size=BATCHSIZE,
        class_mode='categorical')
validation_generator = test_datagen.flow_from_directory(
        './CUB_200_2011/splitted/valid/',
        target_size=TARGET_SIZE,
        batch_size=BATCHSIZE,
        shuffle=False,
        class_mode='categorical')


# In[5]:


def outer_product(x):
    """
    calculate outer-products of 2 tensors

        args 
            x
                list of 2 tensors
                , assuming each of which has shape = (size_minibatch, total_pixels, size_filter)
    """
    return keras.backend.batch_dot(
                x[0]
                , x[1]
                , axes=[1,1]
            ) / x[0].get_shape().as_list()[1] 

def signed_sqrt(x):
    """
    calculate element-wise signed square root

        args
            x
                a tensor
    """
    return keras.backend.sign(x) * keras.backend.sqrt(keras.backend.abs(x) + 1e-9)

def L2_norm(x, axis=-1):
    """
    calculate L2-norm

        args 
            x
                a tensor
    """
    return keras.backend.l2_normalize(x, axis=axis)


# In[6]:


flg_debug=True
keras.backend.clear_session()
img=Input(shape=INPUT_SIZE, name='img')
x=BatchNormalization()(img)
# base_model = VGG16( input_tensor=img, include_top=False,weights='imagenet')
xcep_model = load_model('./xception_triplet_tuned-diff-bigdata1.h5')
# base_model = Xception(include_top=True, weights='imagenet', input_tensor=None, input_shape=None)


# In[7]:


# for i,layer in enumerate(xcep_model.layers):
#     print(i,layer.name)


# In[8]:


base_model = xcep_model.layers[3]
# base_model.layers.pop()
base_model.layers.pop()
base_model.outputs = [base_model.layers[-1].output]
base_model.layers[-1].outbound_nodes = []
base_model.output_layers = [base_model.layers[-1]]
for layer in base_model.layers[:88]:
    layer.trainable=False
for layer in base_model.layers[88:]:
    layer.trainable=True
x=base_model(x)
x=Conv2D(filters=299,padding='valid',activation='relu',kernel_size=(1,1))(x)
x_detector=x
shape_detector = x_detector.shape.as_list()
if flg_debug:
    print("shape_detector : {}".format(shape_detector))
# extract features from extractor , same with detector for symmetry DxD model
shape_extractor = shape_detector
x_extractor = x_detector
if flg_debug:
    print("shape_extractor : {}".format(shape_extractor))
# rehape to (minibatch_size, total_pixels, filter_size)
x_detector = keras.layers.Reshape(
        [
            shape_detector[1] * shape_detector[2] , shape_detector[-1]
        ]
    )(x_detector)
if flg_debug:
    print("x_detector shape after rehsape ops : {}".format(x_detector.shape))

x_extractor = keras.layers.Reshape(
        [
            shape_extractor[1] * shape_extractor[2] , shape_extractor[-1]
        ]
    )(x_extractor)
if flg_debug:
    print("x_extractor shape after rehsape ops : {}".format(x_extractor.shape))


# outer products of features, output shape=(minibatch_size, filter_size_detector*filter_size_extractor)
x = keras.layers.Lambda(outer_product)(
    [x_detector, x_extractor]
)

if flg_debug:
    print("x shape after outer products ops : {}".format(x.shape))

# rehape to (minibatch_size, filter_size_detector*filter_size_extractor)
x = keras.layers.Reshape([shape_detector[-1]*shape_extractor[-1]])(x)
if flg_debug:
    print("x shape after rehsape ops : {}".format(x.shape))

# signed square-root 
x = keras.layers.Lambda(signed_sqrt)(x)
if flg_debug:
    print("x shape after signed-square-root ops : {}".format(x.shape))

# L2 normalization
x = keras.layers.Lambda(L2_norm)(x)
if flg_debug:
    print("x shape after L2-Normalization ops : {}".format(x.shape))
### 
### attach FC-Layer
###
##########################################################################################
category_predict = Dense(200, activation='softmax', name='ctg_out',
                         kernel_initializer=keras.initializers.glorot_normal()
                        )(x)
savepath='./fine_tune.h5'
model = Model(inputs=img, outputs=category_predict)
# model.load_weights('./test.h5')
#定义一下优化的方法
opt = keras.optimizers.SGD(lr=1., momentum=0.9, decay=0.0, nesterov=False)
#compile一下模型
model.compile(loss='categorical_crossentropy',optimizer=opt,metrics=['accuracy'])
print(model.summary())
#train model
callback_reducer = keras.callbacks.ReduceLROnPlateau( monitor='val_loss', factor=0.5, patience=2, min_lr=1e-6)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              
earlyStopping=kcallbacks.EarlyStopping(monitor='val_acc', patience=6, verbose=1, mode='auto')
saveBestModel = kcallbacks.ModelCheckpoint(filepath=savepath, monitor='val_acc', verbose=1, save_best_only=True, mode='auto',save_weights_only=True)
hist=model.fit_generator(
    train_generator,
    steps_per_epoch=666,
    epochs=50,
    validation_data=validation_generator,
    validation_steps=len(validation_generator),
    callbacks=[earlyStopping,saveBestModel,callback_reducer]
)


# In[9]:


# for i,layer in enumerate(base_model.layers):
#     print(i,layer.name)
# for layer in base_model.layers[:88]:
#     layer.trainable=False
# for layer in base_model.layers[88:]:
#     layer.trainable=True


# In[10]:


# #定义一下优化的方法
# opt = keras.optimizers.SGD(lr=1e-3, momentum=0.9, decay=1e-9, nesterov=False)
# #compile一下模型
# model.compile(loss='categorical_crossentropy',optimizer=opt,metrics=['accuracy'])
# print(model.summary())
# #train model
# callback_reducer = keras.callbacks.ReduceLROnPlateau( monitor='val_loss', factor=0.5, patience=5, min_lr=1e-6)  
# earlyStopping=kcallbacks.EarlyStopping(monitor='val_acc', patience=3, verbose=1, mode='auto')
# saveBestModel = kcallbacks.ModelCheckpoint(filepath=savepath, monitor='val_acc', verbose=1, save_best_only=True, 
#                                            save_weights_only=False,
#                                            mode='auto')
# hist=model.fit_generator(
#     train_generator,
#     steps_per_epoch=300,
#     epochs=33,
#     validation_data=validation_generator,
#     validation_steps=5794/BATCHSIZE+1,
#     callbacks=[earlyStopping,saveBestModel,callback_reducer]
# )


# # 进行测试

# In[11]:


print(len(validation_generator))


# In[12]:


loss,acc=model.evaluate_generator(validation_generator,steps=len(validation_generator))


# In[13]:


print('测试集的loss是:{:.2f},acc是：{:.2f}'.format(loss,acc))

