
# coding: utf-8

# In[1]:


import os
import numpy as np
from keras import Input
from keras.applications import Xception, InceptionV3
from keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau
from keras.layers import Dense, Dropout, concatenate, maximum,Conv2D,Reshape,GlobalAveragePooling2D
from keras.models import Model,load_model
from keras.optimizers import SGD
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import plot_model,to_categorical
from keras.layers.core import Lambda
import os
import keras
from collections import OrderedDict
from sklearn.preprocessing import LabelEncoder
import random
from keras.preprocessing import image
from numpy.random import randint, shuffle, choice


# In[2]:


#写一个LossHistory类，保存loss和acc
class LossHistory(keras.callbacks.Callback):
    def on_train_begin(self, logs={}):
        self.losses = {'batch':[], 'epoch':[]}
        self.accuracy = {'batch':[], 'epoch':[]}
        self.val_loss = {'batch':[], 'epoch':[]}
        self.val_acc = {'batch':[], 'epoch':[]}
    def on_batch_end(self, batch, logs={}):
        self.losses['batch'].append(logs.get('loss'))
        self.accuracy['batch'].append(logs.get('acc'))
        self.val_loss['batch'].append(logs.get('val_loss'))
        self.val_acc['batch'].append(logs.get('val_acc'))
    def on_epoch_end(self, batch, logs={}):
        self.losses['epoch'].append(logs.get('loss'))
        self.accuracy['epoch'].append(logs.get('acc'))
        self.val_loss['epoch'].append(logs.get('val_loss'))
        self.val_acc['epoch'].append(logs.get('val_acc'))
    def loss_plot(self, loss_type):
        iters = range(len(self.losses[loss_type]))
        plt.figure()
        plt.plot(iters, self.losses[loss_type], 'g', label='train loss')
        if loss_type == 'epoch':
            # acc
            plt.plot(iters, self.accuracy[loss_type], 'r', label='train acc')
            # loss
            plt.plot(iters, self.losses[loss_type], 'g', label='train loss')
            # val_acc
            plt.plot(iters, self.val_acc[loss_type], 'b', label='val acc')
            # val_loss
            plt.plot(iters, self.val_loss[loss_type], 'k', label='val loss')
        plt.grid(True)
        plt.xlabel(loss_type)
        plt.ylabel('acc-loss')
        plt.legend(loc="upper right")
        plt.show()


# In[3]:


#创建一个实例来存储history
local_history= LossHistory()


# In[4]:


#定义一些基本的变量
BATCH_SIZE=48


# In[5]:


labels=[]
for name in os.listdir('./CUB_200_2011/299/train/'):
    labels.append(name)
print(len(labels))


# In[6]:


le=LabelEncoder()
le.fit(labels)


# In[7]:


rootdir = './CUB_200_2011/299/train/'
train = OrderedDict()
for subdir, dirs, files in os.walk(rootdir):
    for file in files:
        if not file.endswith(".jpg"):
            continue
        label=le.transform([subdir.split('/')[-1]])[0]
        if label not in train:
            train[label] = []
        train[label].append(os.path.join(subdir, file))


# In[8]:


len(train)


# In[9]:


def triplet_generator(class_img_labels, batch_size, train=True):
    while True:
        mid_label = randint(len(class_img_labels), size=batch_size)
        left_label=np.copy(mid_label)
        right_label = np.copy(mid_label)
        shuffle(right_label)
        # select by label
        mid_images=list()
        left_images = list()
        right_images = list()
        for i in range(batch_size):
            mid_images.append(image.img_to_array(image.load_img(random.choice(class_img_labels[mid_label[i]])))/255)
            left_images.append(image.img_to_array(image.load_img(random.choice(class_img_labels[left_label[i]])))/255)
            right_images.append(image.img_to_array(image.load_img(random.choice(class_img_labels[right_label[i]])))/255)
        mid_images=np.array(mid_images)
        left_images = np.array(left_images)
        right_images = np.array(right_images)
        binary_label_left= to_categorical((mid_label == left_label).astype(int),num_classes=2)
        binary_label_right= to_categorical((mid_label == right_label).astype(int),num_classes=2)
        mid_label = to_categorical(mid_label, num_classes=len(class_img_labels))
        left_label = to_categorical(left_label, num_classes=len(class_img_labels))
        right_label = to_categorical(right_label, num_classes=len(class_img_labels))
        yield [mid_images,left_images, right_images], [mid_label,left_label, right_label, binary_label_left,binary_label_right]


# In[10]:


rootdir = './CUB_200_2011/299/valid/'
test = OrderedDict()
for subdir, dirs, files in os.walk(rootdir):
    for file in files:
        if not file.endswith(".jpg"):
            continue
        label=le.transform([subdir.split('/')[-1]])[0]
        if label not in test:
            test[label] = []
        test[label].append(os.path.join(subdir, file))


# In[11]:


len(test)


# In[12]:


# next(triplet_generator(train,batch_size=2))


# In[13]:


# train_datagen = ImageDataGenerator(rescale=1. / 255)
# test_datagen = ImageDataGenerator(rescale=1. / 255)
# train_generator = train_datagen.flow_from_directory(
#     '../BirdAPP/',
#     target_size=(256, 256),
#     batch_size=BATCH_SIZE,
#     class_mode='categorical')
# validation_generator = test_datagen.flow_from_directory(
#     '../test/',
#     # '/home/cwh/coding/data/cwh/test1',
#     target_size=(256, 256),
#     batch_size=BATCH_SIZE,
#     class_mode='categorical')
# def pair_generator(cur_generator, batch_size, train=True):
#     cur_cnt = 0
#     while True:
#         if train and cur_cnt % 4 == 1:
#             # provide same image
#             x1, y1 = train_generator.next()
#             if y1.shape[0] != batch_size:
#                 x1, y1 = train_generator.next()
#             # print(y1)
#             # print(np.sort(np.argmax(y1, 1), 0))
#             y1_labels = np.argmax(y1, 1)
#             has_move = list()
#             last_not_move = list()
#             idx2 = [-1 for i in range(batch_size)]

#             for i, label in enumerate(y1_labels):
#                 if i in has_move:
#                     continue
#                 for j in range(i+1, batch_size):
#                     if y1_labels[i] == y1_labels[j]:
#                         idx2[i] = j
#                         idx2[j] = i
#                         has_move.append(i)
#                         has_move.append(j)
#                         break
#                 if idx2[i] == -1:
#                     # same element not found and hasn't been moved
#                     if len(last_not_move) == 0:
#                         last_not_move.append(i)
#                         idx2[i] = i
#                     else:
#                         idx2[i] = last_not_move[-1]
#                         idx2[last_not_move[-1]] = i
#                         del last_not_move[-1]
#             x2 = list()
#             y2 = list()
#             for i2 in range(batch_size):
#                 x2.append(x1[idx2[i2]])
#                 y2.append(y1[idx2[i2]])
#             # print(y2)
#             x2 = np.asarray(x2)
#             y2 = np.asarray(y2)
#             # print(x2.shape)
#             # print(y2.shape)
#         else:
#             x1, y1 = cur_generator.next()
#             if y1.shape[0] != batch_size:
#                 x1, y1 = cur_generator.next()
#             x2, y2 = cur_generator.next()
#             if y2.shape[0] != batch_size:
#                 x2, y2 = cur_generator.next()
#         same = (np.argmax(y1, 1) == np.argmax(y2, 1)).astype(int)
#         one_hot_same = np.zeros([batch_size, 2])
#         one_hot_same[np.arange(batch_size), same] = 1
#         # print same
#         # print one_hot_same
#         # print(np.argmax(y1, 1))
#         # print(np.argmax(y2, 1))
#         # print(same)
#         cur_cnt += 1
#         yield [x1, x2], [y1, y2, one_hot_same]


# In[14]:


def eucl_dist(inputs):
    x, y = inputs
    return (x - y)**2
def outer_product(x):
    """
    calculate outer-products of 2 tensors

        args 
            x
                list of 2 tensors
                , assuming each of which has shape = (size_minibatch, total_pixels, size_filter)
    """
    return keras.backend.batch_dot(
                x[0]
                , x[1]
                , axes=[1,1]
            ) / x[0].get_shape().as_list()[1] 

def signed_sqrt(x):
    """
    calculate element-wise signed square root

        args
            x
                a tensor
    """
    return keras.backend.sign(x) * keras.backend.sqrt(keras.backend.abs(x) + 1e-9)

def L2_norm(x, axis=-1):
    """
    calculate L2-norm

        args 
            x
                a tensor
    """
    return keras.backend.l2_normalize(x, axis=axis)


# In[15]:


early_stopping = EarlyStopping(monitor='val_loss', patience=3)
auto_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=3, verbose=0, mode='auto', epsilon=0.0001, cooldown=0, min_lr=0)
save_model = ModelCheckpoint('xception_triplet.h5',save_best_only=True,save_weights_only=False)


# In[16]:


if os.path.exists('xception-tuned-diff.h5'):
    print(111)
    model = load_model('xception_triplet.h5')
else:
    # create the base pre-trained model
    input_tensor = Input(shape=(299, 299, 3))
    xcep_model = load_model('./xception_triplet_tuned-diff-bigdata.h5')
    base_model = xcep_model.layers[3]
#     base_model = Xception(include_top=True, weights='imagenet', input_tensor=None, input_shape=None)
    plot_model(base_model, to_file='xception_model.png')
    base_model.layers.pop()
#     base_model.layers.pop()
    base_model.outputs = [base_model.layers[-1].output]
    base_model.layers[-1].outbound_nodes = []
    base_model.output_layers = [base_model.layers[-1]]

    feature = base_model
    for layer in base_model.layers[:88]:
        layer.trainable=False
    for layer in base_model.layers[88:]:
        layer.trainable=True
    img1 = Input(shape=(299, 299, 3), name='img_1')
    img2 = Input(shape=(299, 299, 3), name='img_2')
    img3 = Input(shape=(299, 299, 3), name='img_3')
    glob=GlobalAveragePooling2D()
    feature1 = feature(img1)
    feature2 = feature(img2)
    feature3=  feature(img3)
    reduce_dimension=Conv2D(filters=100,padding='valid',activation='relu',kernel_size=(1,1))
    reshape_1=Reshape([10*10,100])
    outproduct=keras.layers.Lambda(outer_product)
    reshape_2=Reshape([100*100])
    sign=keras.layers.Lambda(signed_sqrt)
    feature_1 = reshape_1(reduce_dimension(feature1))
    feature_2 = reshape_1(reduce_dimension(feature2))
    feature_3=  reshape_1(reduce_dimension(feature3))
    feature_1=sign(reshape_2(outproduct([feature_1,feature_1])))
    feature_2=sign(reshape_2(outproduct([feature_2,feature_2])))
    feature_3=sign(reshape_2(outproduct([feature_3,feature_3])))               
    # let's add a fully-connected layer
    category_predict1 = Dense(200, activation='softmax', name='ctg_out_1')(
        (feature_1)
    )
    category_predict2 = Dense(200, activation='softmax', name='ctg_out_2')(
       (feature_2)
    )
    category_predict3=Dense(200, activation='softmax', name='ctg_out_3')(
        (feature_3)
    )
    # concatenated = keras.layers.concatenate([feature1, feature2])
    feature1 = glob(feature(img1))
    feature2 = glob(feature(img2))
    feature3=  glob(feature(img3))
    dis_1 = Lambda(eucl_dist, name='square1')([feature1, feature2])
    dis_2 = Lambda(eucl_dist,name='square2')([feature1, feature3])
    # concatenated = Dropout(0.5)(concatenated)
    # let's add a fully-connected layer
    # x = Dense(1024, activation='relu')(concatenated)
    out=Dense(2,activation='softmax',name='pair')
    judge1 = out(dis_1)
    judge2 = out(dis_2)
    
    model = Model(inputs=[img1,img2,img3], outputs=[category_predict1, category_predict2,category_predict3,judge1,judge2])

    # model.save('dog_xception.h5')
    plot_model(model, to_file='model_triplet.png')
    # first: train only the top layers (which were randomly initialized)
    # i.e. freeze all convolutional InceptionV3 layers
    for layer in base_model.layers:
        layer.trainable = False

    # compile the model (should be done *after* setting layers to non-trainable)
    model.compile(optimizer='nadam',
                  loss={'ctg_out_1': 'categorical_crossentropy',
                        'ctg_out_2': 'categorical_crossentropy',
                        'ctg_out_3':'categorical_crossentropy',
                        'pair':'categorical_crossentropy',
#                         'pairsame': 'binary_crossentropy',
#                         'pairdiff':'binary_crossentropy'
                       },
                  loss_weights={
                      'ctg_out_1': 1/3,
                      'ctg_out_2':1/3,
                      'ctg_out_3':1/3,
                      'pair':0,
                  },
                  metrics=['accuracy'])
    # model = make_parallel(model, 3)
    # train the model on the new data for a few epochs
    model.fit_generator(triplet_generator(train, batch_size=BATCH_SIZE),
                        steps_per_epoch=40000/BATCH_SIZE+1,
                        epochs=15,
                        validation_data=triplet_generator(test,batch_size=BATCH_SIZE),
                        validation_steps=5480/BATCH_SIZE+1,
                        callbacks=[early_stopping, auto_lr, save_model])
    model.save('xception_triplet.h5')


# In[17]:


for i, layer in enumerate(model.layers[3].layers):
    print(i, layer.name)


# In[ ]:


# we chose to train the top 2 inception blocks, i.e. we will freeze
# the first 172 layers and unfreeze the rest:
cur_base_model = model.layers[3]
for layer in cur_base_model.layers[:88]:
    layer.trainable = False
for layer in cur_base_model.layers[88:]:
    layer.trainable = True
# we need to recompile the model for these modifications to take effect
# we use SGD with a low learning rate

from keras.optimizers import SGD
model.compile(
            optimizer=SGD(lr=0.0001, momentum=0.9),
#             optimizer='adam',
            loss={      'ctg_out_1': 'categorical_crossentropy',
                        'ctg_out_2': 'categorical_crossentropy',
                        'ctg_out_3':'categorical_crossentropy',
                        'pair': 'categorical_crossentropy',
                       },
                  loss_weights={
                      'ctg_out_1': 1/3,
                      'ctg_out_2': 1/3,
                      'ctg_out_3':1/3,
                      'pair':1},
              metrics=['accuracy']
              )

# we train our model again (this time fine-tuning the top 2 inception blocks
# alongside the top Dense layers
save_model = ModelCheckpoint('xception-tuned-diff.h5',save_best_only=True,save_weights_only=False)
model.fit_generator(triplet_generator(train, batch_size=BATCH_SIZE),
                    steps_per_epoch=40000/BATCH_SIZE+1,
                    epochs=20,
                    validation_data=triplet_generator(test,batch_size=BATCH_SIZE),
                    validation_steps=5480/BATCH_SIZE+1,
                    callbacks=[auto_lr, save_model,local_history]) # otherwise the generator would loop indefinitely
model.save('xception_triplet_tuned-diff.h5')


# In[ ]:


import matplotlib.pyplot as plt
import pickle
to_save={}
to_save['losses']=local_history.losses
to_save['accuracy']=local_history.accuracy
to_save['val_loss']=local_history.val_loss
to_save['val_acc']=local_history.val_acc
with open('local_history.pickle','wb') as f:
    pickle.dump(to_save,f)
local_history.loss_plot('epoch')
local_history.loss_plot('batch')
plt.show()

